package com.sillycycle.bagleyd.dino;

/*
 * @(#)Dino3D.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import javax.swing.JFrame;

/**
 * The <code>Dino3D</code> class does all the 3D graphics for
 * <code>DinoFrame</code> class.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class Dino3D extends Canvas {
	private static final long serialVersionUID = 42L;
	static final int SAME = 0;
	static final int OPPOSITE = 1;
	static final int DOWN = 0;
	static final int UP = 1;
	static final int COORD3D = 12;
	static final int MAX_VIEWS = 2;

	static final double SQRT_3 = 1.732050808;

	static int multiply(int a) {
		return ((int) (a *  SQRT_3 / 2.0));
	}

	static int divide(int a) {
		return ((int) (a * 2.0 / SQRT_3));
	}

	static final int[/*MAX_RECT*/] planeToCube =
		{6, 0, 6, 1, 2, 3, 6, 4, 6, 6, 5, 6};
	static int[/*MAX_FACES*/] cubeToPlane =
		{1, 3, 4, 5, 7, 10};

	Point pos = new Point(0, 0);
	Point puzzleSize, viewSize, cubeSize, faceSize, coreSize;
	Point puzzleOffset, letterOffset, viewMiddle;
	int faceDiagonal, cubeDiagonal, delta;
	int lastX, lastY;

	Point[][] faceLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT];
	Point[][] cubeLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT];
	Point[][] diamondLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT + 1];
	Point[][] tinyDiamondLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT + 1];
	Point[][][] triangleLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT][4];
	Point[][] letter3DList = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT];
	Point[][][] orientTriangle = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT][2];
	Point start = new Point(0, 0);

	JFrame      frame = null;
	Dino       dino = null;
	int         randomPaints = 0;
	boolean     firstPaint = false, resizePaint = false;
	boolean     scanPaint = false, stackPaint = false;
	boolean     framePaint = false, allPiecesPaint = false;
	boolean     selectPaint = false, releasePaint = false;
	boolean     mouseDown = false, vertical = false;

	public Dino3D(JFrame frame, Dino dino) {
		this.frame = frame;
		this.dino = dino;
		firstPaint = true;
	}

	public boolean checkStarted() {
		return dino.checkStarted();
	}

	Point diamondOffset3D(int face) {
		if (vertical) { // return starting pt
			switch (face) {
			case 0:
			  return new Point(
			    (2 * faceLoc3D[0][0].x + faceLoc3D[0][1].x +
			      faceLoc3D[0][2].x) / 2,
			    (2 * faceLoc3D[0][0].y +
			      faceLoc3D[0][1].y) / 2);
			case 1:
			  return new Point(
			    faceLoc3D[1][0].x + faceLoc3D[1][1].x,
			    faceLoc3D[1][0].y);
			case 2:
			  return new Point(
			    (2 * faceLoc3D[2][0].x + faceLoc3D[2][1].x +
			      faceLoc3D[2][2].x) / 2,
			    (2 * faceLoc3D[2][0].y + faceLoc3D[2][1].y +
			      faceLoc3D[2][2].y) / 2);
			case 3:
			  return new Point(
			    faceLoc3D[3][0].x + faceLoc3D[3][1].x,
			    faceLoc3D[3][0].y);
			case 4:
			  return new Point(
			    (2 * faceLoc3D[4][0].x + faceLoc3D[4][1].x +
			      faceLoc3D[4][2].x) / 2,
			    (2 * faceLoc3D[4][0].y + faceLoc3D[4][1].y +
			      faceLoc3D[4][2].y) / 2);
			case 5:
			  return new Point(
			    (2 * faceLoc3D[5][0].x + faceLoc3D[5][1].x +
			      faceLoc3D[5][2].x) / 2,
			    (2 * faceLoc3D[5][0].y +
			      faceLoc3D[5][1].y) / 2);
			default:
			  System.err.println("diamondOffset3D: face " +
			    face);
			}
		} else {
			switch (face) {
			case 0:
			  return new Point(
			    faceLoc3D[0][0].x,
			    faceLoc3D[0][0].y + faceLoc3D[0][1].y);
			case 1:
			  return new Point(
			    (2 * faceLoc3D[1][0].x + faceLoc3D[1][1].x +
			      faceLoc3D[1][2].x) / 2,
			    (2 * faceLoc3D[1][0].y + faceLoc3D[1][1].y +
			      faceLoc3D[1][2].y) / 2);
			case 2:
			  return new Point(
			    (2 * faceLoc3D[2][0].x +
			      faceLoc3D[2][1].x) / 2,
			    (2 * faceLoc3D[2][0].y + faceLoc3D[2][1].y +
			      faceLoc3D[2][2].y) / 2);
			case 3:
			  return new Point(
			    (2 * faceLoc3D[3][0].x +
			      faceLoc3D[3][1].x) / 2,
			    (2 * faceLoc3D[3][0].y + faceLoc3D[3][1].y +
			      faceLoc3D[3][2].y) / 2);
			case 4:
			  return new Point(
			    faceLoc3D[4][0].x,
			    faceLoc3D[4][0].y + faceLoc3D[4][1].y);
			case 5:
			  return new Point(
			    (2 * faceLoc3D[5][0].x + faceLoc3D[5][1].x +
			      faceLoc3D[5][2].x) / 2,
			    (2 * faceLoc3D[5][0].y + faceLoc3D[5][1].y +
			      faceLoc3D[5][2].y) / 2);
			default:
			  System.err.println("diamondOffset3D: face " +
			    face);
			}
		}
		return new Point(0, 0);
	}

	static int mapTo3D(int face, int side) {
		switch (face) { // return corner
		case 0:
			return (side + 2) % Dino.MAX_ORIENT;
		case 1:
		case 5:
			return side;
		case 2:
			return (side + 1) % Dino.MAX_ORIENT;
		case 3:
		case 4:
			return (side + 3) % Dino.MAX_ORIENT;
		default:
			System.err.println("mapTo3D: face " + face);
		}
		return -1;
	}

	static int mapFrom3D(int face, int corner) {
		switch (face) { // return side
		case 0:
			return (corner + 2) % Dino.MAX_ORIENT;
		case 1:
		case 5:
			return corner;
		case 2:
			return (corner + 3) % Dino.MAX_ORIENT;
		case 3:
		case 4:
			return (corner + 1) % Dino.MAX_ORIENT;
		default:
			System.err.println("mapFrom3D: face " + face);
		}
		return -1;
	}

	Point cubeOffset3D(int face, int side) {
		Point point;

		point = diamondOffset3D(face);
		for (int corner = 1; corner <= side; corner++) {
			point.x += tinyDiamondLoc3D[face][corner].x;
			point.y += tinyDiamondLoc3D[face][corner].y;
		}
		return point;
	}

	static void fillPolygon(Graphics g, Color fillColor, Color lineColor,
			int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.setColor(fillColor);
		g.fillPolygon(tempListX, tempListY, n);
		g.setColor(lineColor);
		g.drawPolygon(tempListX, tempListY, n);
	}


	static void drawPolygon(Graphics g, int startX, int startY,
			int[] listX, int[] listY, int n) {
		int[] tempListX = new int[n];
		int[] tempListY = new int[n];

		for (int i = 0; i < n; i++) {
			tempListX[i] = listX[i] + startX;
			tempListY[i] = listY[i] + startY;
		}
		g.drawPolygon(tempListX, tempListY, n);
	}

	void drawTriangle(int face, int position, int offsetPosition) {
		Graphics g = dino.getGraphics();

		if (g == null)
			return;
		try {
			int corner;
			int rotation, l, letterX, letterY;
			Point startCoord;
			Color fillColor, lineColor;

			corner = mapTo3D(face, position);
			startCoord = cubeOffset3D(face, corner);
			rotation = dino.cubeLoc[face][position].rotation;
			letterX = startCoord.x + letter3DList[face][corner].x;
			letterY = startCoord.y + letter3DList[face][corner].y;
			/* triangleLoc3D[face][corner][0].x = start.x;
			triangleLoc3D[face][corner][0].y = start.y; */
			l = dino.cubeLoc[face][position].face;
			if (offsetPosition == 0) {
				if (dino.mono) {
					if (dino.reverse) {
						lineColor = Color.black;
						fillColor = Color.white;
					} else {
						lineColor = Color.white;
						fillColor = Color.black;
					}
				} else {
					fillColor = dino.faceColor[l];
					lineColor = dino.borderColor;
				}
			} else {
				if (dino.mono) {
					if (dino.reverse) {
						lineColor = Color.white;
						fillColor = Color.black;
					} else {
						lineColor = Color.black;
						fillColor = Color.white;
					}
				} else {
					lineColor = dino.faceColor[l];
					fillColor = dino.borderColor;
				}
			}
			int[] tempListX = new int[3];
			int[] tempListY = new int[3];

			for (int k = 0; k < 3; k++) {
				tempListX[k] =
					triangleLoc3D[face][corner][k].x +
					((k > 0) ?
					tempListX[k - 1] : 0);
				tempListY[k] =
					triangleLoc3D[face][corner][k].y +
					((k > 0) ?
					tempListY[k - 1] : 0);
			}
			fillPolygon(g, fillColor, lineColor,
				startCoord.x, startCoord.y,
				tempListX, tempListY, 3);
			if (dino.mono) {
				String buf;
				int fontSize;

				g.setColor(lineColor);
				fontSize = cubeDiagonal / 4;
				g.setFont(new Font("arial", 0, fontSize));

				buf = Character.toString(dino.faceNames.charAt(l));
				g.drawString(buf, letterX - 2, letterY + 2);
			}
			if (dino.orientChoice) {
				g.setColor(lineColor);
				g.drawLine(letterX +
				  orientTriangle[face][rotation][0].x,
				  letterY +
				  orientTriangle[face][rotation][0].y,
				  letterX +
				  orientTriangle[face][rotation][1].x,
				  letterY +
				  orientTriangle[face][rotation][1].y);
			}
		} finally {
			g.dispose();
		}
	}

	void drawAllPieces() {
		for (int face = 0; face < Dino.MAX_FACES; face++)
			drawFace(face);
	}

	void drawFace(int face) {
		for (int position = 0; position < Dino.MAX_ORIENT; position++)
			drawTriangle(face, position, 0);
	}

	void drawFrame(boolean show) {
		Graphics g = dino.getGraphics();

		if (g == null)
			return;
		try {
			if (show) {
				Point coord = new Point(viewSize.x + puzzleOffset.x,
					viewSize.y + puzzleOffset.y);
				int fontSize = cubeDiagonal / 3;

				g.setFont(new Font("arial", 0, fontSize));
				g.setColor(dino.foreground);
				if (vertical) {
					g.drawLine(0, coord.y,
					  coord.x + puzzleOffset.x + 1,
					  coord.y);
					g.drawString("Front", 2 * delta,
					  3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
					  -3 * delta - 10 * fontSize / 4 + coreSize.x,
					  -delta - 2 * fontSize / 5 + coreSize.y);
				} else {
					g.drawLine(coord.x, 0,
					  coord.x, coord.y + puzzleOffset.y + 1);
					g.drawString("Front", 2 * delta,
					  3 * delta + 2 * fontSize / 5);
					g.drawString("Back",
					  -3 * delta - 10 * fontSize / 4 + coreSize.x,
					  -delta - 2 * fontSize / 5 + coreSize.y);
				}
				g.setColor(dino.currentForeground);
				for (int face = 0; face < Dino.MAX_FACES; face++) {
					int[] tempListX = new int[Dino.MAX_ORIENT];
					int[] tempListY = new int[Dino.MAX_ORIENT];

					for (int i = 0; i < Dino.MAX_ORIENT; i++) {
						tempListX[i] =
							faceLoc3D[face][i].x +
							((i > 0) ?
							tempListX[i - 1] : 0);
						tempListY[i] =
							faceLoc3D[face][i].y +
							((i > 0) ?
							tempListY[i - 1] : 0);
					}
					drawPolygon(g, 0, 0,
						tempListX, tempListY,
						Dino.MAX_ORIENT);
				}
			} else {
				/*g.setColor(Color.white);
				g.fillRect(0, 0, coreSize.x, offsetY);
				g.fillRect(0, coreSize.y - offsetY - 1,
					coreSize.x, offsetY + 1);
				g.fillRect(0, 0, offsetX, coreSize.y);
				g.fillRect(coreSize.x - offsetX - 1, 0,
					offsetX + 1, coreSize.y);
				g.setColor(background);
				g.fillRect(puzzleOffset.x + 1,
					puzzleOffset.y + 1,
					coreSize.x - 2 - 2 * puzzleOffset.x,
					coreSize.y - 2 - 2 * puzzleOffset.y);*/
				g.setColor(dino.background);
				g.fillRect(0, 0, coreSize.x, coreSize.y);
			}
		} finally {
			g.dispose();
		}
	}

	void resizePieces() {
		int side, corner, orient;
		Point[][] subcubeLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT];
		Point[][][] orientCubeLoc3D = new Point[2][Dino.MAX_FACES][Dino.MAX_ORIENT];
		Point[][] subdiamondLoc3D = new Point[Dino.MAX_FACES][Dino.MAX_ORIENT];
		Point[][][] orientDiamondLoc3D = new Point[2][Dino.MAX_FACES][Dino.MAX_ORIENT];
		/* letterOffset.x = -2;
		letterOffset.y = 3; */
		viewMiddle = new Point(faceSize.x + puzzleOffset.x,
			faceSize.y + puzzleOffset.y);
		for (int face = 0; face < Dino.MAX_FACES; face++) {
			faceLoc3D[face][0] = new Point(viewMiddle.x,
				viewMiddle.y);
			for (orient = 1; orient < Dino.MAX_ORIENT; orient++) {
				faceLoc3D[face][orient] = new Point(faceSize.x,
					faceSize.y);
			}
		}
		if (vertical) {
			faceLoc3D[0][1].x /= -2;
			faceLoc3D[0][1].y /= -1;
			faceLoc3D[0][2].y = 0;
			faceLoc3D[0][3].x /= 2;

			faceLoc3D[1][1].x /= -2;
			faceLoc3D[1][2].x /= -2;
			faceLoc3D[1][2].y /= -1;
			faceLoc3D[1][3].x /= 2;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y = 0;
			faceLoc3D[2][2].x /= -2;
			faceLoc3D[2][3].x /= -1;
			faceLoc3D[2][3].y = 0;

			for (int face = Dino.MAX_FACES / 2; face < Dino.MAX_FACES; face++)
				faceLoc3D[face][0].y += viewSize.y + 3;

			faceLoc3D[3][1].x /= 2;
			faceLoc3D[3][1].y /= -1;
			faceLoc3D[3][2].x /= 2;
			faceLoc3D[3][3].x /= -2;

			faceLoc3D[4][1].x /= -1;
			faceLoc3D[4][1].y = 0;
			faceLoc3D[4][2].x /= 2;
			faceLoc3D[4][2].y /= -1;
			faceLoc3D[4][3].y = 0;

			faceLoc3D[5][1].x /= 2;
			faceLoc3D[5][2].x /= -1;
			faceLoc3D[5][2].y = 0;
			faceLoc3D[5][3].x /= -2;
			faceLoc3D[5][3].y /= -1;
		} else {
			faceLoc3D[0][1].x /= -1;
			faceLoc3D[0][1].y /= -2;
			faceLoc3D[0][2].y /= -2;
			faceLoc3D[0][3].y /= 2;

			faceLoc3D[1][1].x = 0;
			faceLoc3D[1][2].x /= -1;
			faceLoc3D[1][2].y /= -2;
			faceLoc3D[1][3].x = 0;
			faceLoc3D[1][3].y /= -1;

			faceLoc3D[2][1].y /= -2;
			faceLoc3D[2][2].x = 0;
			faceLoc3D[2][3].y /= 2;
			faceLoc3D[2][3].x /= -1;

			for (int face = Dino.MAX_FACES / 2; face < Dino.MAX_FACES; face++)
				faceLoc3D[face][0].x += viewSize.x + 3;

			faceLoc3D[3][1].x /= -1;
			faceLoc3D[3][1].y /= 2;
			faceLoc3D[3][2].x = 0;
			faceLoc3D[3][2].y /= -1;
			faceLoc3D[3][3].y /= -2;

			faceLoc3D[4][1].y /= 2;
			faceLoc3D[4][2].x /= -1;
			faceLoc3D[4][2].y /= 2;
			faceLoc3D[4][3].x /= -1;
			faceLoc3D[4][3].y /= -2;

			faceLoc3D[5][1].x = 0;
			faceLoc3D[5][1].y /= -1;
			faceLoc3D[5][2].y /= 2;
			faceLoc3D[5][3].x = 0;
		}
		for (int face = 0; face < Dino.MAX_FACES; face++) {
			cubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			subcubeLoc3D[face][0] = new Point(faceLoc3D[face][0].x,
				faceLoc3D[face][0].y);
			orientCubeLoc3D[0][face][0] = new Point(0, 0);
			orientCubeLoc3D[1][face][0] = new Point(0, 0);
			for (orient = 1; orient < Dino.MAX_ORIENT; orient++) {
				cubeLoc3D[face][orient] = new Point(
					faceLoc3D[face][orient].x - 3 *
					delta * faceLoc3D[face][orient].x /
					faceSize.x,
					faceLoc3D[face][orient].y - 3 *
					delta * faceLoc3D[face][orient].y /
					faceSize.y);
				subcubeLoc3D[face][orient] = new Point(
					faceLoc3D[face][orient].x - 4 *
					delta * faceLoc3D[face][orient].x /
					faceSize.x,
					faceLoc3D[face][orient].y - 4 *
					delta * faceLoc3D[face][orient].y /
					faceSize.y);
				orientCubeLoc3D[0][face][orient] = new Point(
					(faceLoc3D[face][orient].x -
					5 * faceLoc3D[face][orient].x * delta /
					faceSize.x) / 3,
					(faceLoc3D[face][orient].y -
					5 * faceLoc3D[face][orient].y * delta /
					faceSize.y) / 3);
				orientCubeLoc3D[1][face][orient] = new Point(
					(faceLoc3D[face][orient].x -
					7 * faceLoc3D[face][orient].x * delta /
					faceSize.x) / 6,
					(faceLoc3D[face][orient].y -
					7 * faceLoc3D[face][orient].y * delta /
					faceSize.y) / 6);
			}

			for (orient = 0; orient < Dino.MAX_ORIENT; orient++)
				for (int j = 0; j < 4; j++)
					triangleLoc3D[face][orient][j] = new Point(0, 0);

			corner = mapFrom3D(face, 1);
			orientTriangle[face][corner][0] = new Point(
				-orientCubeLoc3D[0][face][2].x / 2,
				-orientCubeLoc3D[0][face][2].y / 2);
			orientTriangle[face][corner][1] = new Point(
				-orientCubeLoc3D[1][face][2].x / 2,
				-orientCubeLoc3D[1][face][2].y / 2);
			for (orient = 0; orient < Dino.MAX_ORIENT - 1; orient++) {
				diamondLoc3D[face][orient] = new Point(
					(cubeLoc3D[face][orient].x +
					cubeLoc3D[face][orient + 1].x) / 2,
					(cubeLoc3D[face][orient].y +
					cubeLoc3D[face][orient + 1].y) / 2);
				subdiamondLoc3D[face][orient] = new Point(
					(subcubeLoc3D[face][orient].x +
					subcubeLoc3D[face][orient + 1].x) / 2,
					(subcubeLoc3D[face][orient].y +
					subcubeLoc3D[face][orient + 1].y) / 2);
				tinyDiamondLoc3D[face][orient + 1] = new Point(
					diamondLoc3D[face][orient].x -
					subdiamondLoc3D[face][orient].x,
					diamondLoc3D[face][orient].y -
					subdiamondLoc3D[face][orient].y);
				orientDiamondLoc3D[0][face][orient] = new Point(
					(orientCubeLoc3D[0][face][orient].x +
					orientCubeLoc3D[0][face][orient + 1].x) / 2,
					(orientCubeLoc3D[0][face][orient].y +
					orientCubeLoc3D[0][face][orient + 1].y) / 2);
				orientDiamondLoc3D[1][face][orient] = new Point(
					(orientCubeLoc3D[1][face][orient].x +
					orientCubeLoc3D[1][face][orient + 1].x) / 2,
					(orientCubeLoc3D[1][face][orient].y +
					orientCubeLoc3D[1][face][orient + 1].y) / 2);
			}

			/* Its a parallelagram so take advantage of that */
			diamondLoc3D[face][orient] = new Point(
				(cubeLoc3D[face][Dino.MAX_ORIENT - 1].x -
				cubeLoc3D[face][Dino.MAX_ORIENT / 2].x) / 2,
				(cubeLoc3D[face][Dino.MAX_ORIENT - 1].y -
				cubeLoc3D[face][Dino.MAX_ORIENT / 2].y) / 2);
			subdiamondLoc3D[face][orient] = new Point(
				(subcubeLoc3D[face][Dino.MAX_ORIENT - 1].x -
				subcubeLoc3D[face][Dino.MAX_ORIENT / 2].x) / 2,
				(subcubeLoc3D[face][Dino.MAX_ORIENT - 1].y -
				subcubeLoc3D[face][Dino.MAX_ORIENT / 2].y) / 2);
			tinyDiamondLoc3D[face][0] = new Point(
				diamondLoc3D[face][orient].x -
				subdiamondLoc3D[face][orient].x,
				diamondLoc3D[face][orient].y -
				subdiamondLoc3D[face][orient].y);
			orientDiamondLoc3D[0][face][orient] = new Point(
				(orientCubeLoc3D[0][face][Dino.MAX_ORIENT - 1].x -
				orientCubeLoc3D[0][face][Dino.MAX_ORIENT / 2].x) / 2,
				(orientCubeLoc3D[0][face][Dino.MAX_ORIENT - 1].y -
				orientCubeLoc3D[0][face][Dino.MAX_ORIENT / 2].y) / 2);
			orientDiamondLoc3D[1][face][orient] = new Point(
				(orientCubeLoc3D[1][face][Dino.MAX_ORIENT - 1].x -
				orientCubeLoc3D[1][face][Dino.MAX_ORIENT / 2].x) / 2,
				(orientCubeLoc3D[1][face][Dino.MAX_ORIENT - 1].y -
				orientCubeLoc3D[1][face][Dino.MAX_ORIENT / 2].y) / 2);

			corner = mapFrom3D(face, 1);
			orientTriangle[face][corner][0] = new Point(
				-orientCubeLoc3D[0][face][2].x / 2,
				-orientCubeLoc3D[0][face][2].y / 2);
			orientTriangle[face][corner][1] = new Point(
				-orientCubeLoc3D[1][face][2].x / 2,
				-orientCubeLoc3D[1][face][2].y / 2);
			triangleLoc3D[face][0][1].x = subdiamondLoc3D[face][2].x;
			triangleLoc3D[face][0][2].x = -subdiamondLoc3D[face][1].x -
				subdiamondLoc3D[face][2].x;
			triangleLoc3D[face][0][1].y = subdiamondLoc3D[face][2].y;
			triangleLoc3D[face][0][2].y = -subdiamondLoc3D[face][1].y -
				subdiamondLoc3D[face][2].y;
			triangleLoc3D[face][1][1].x = -subdiamondLoc3D[face][1].x;
			triangleLoc3D[face][1][2].x = subdiamondLoc3D[face][1].x -
				subdiamondLoc3D[face][2].x;
			triangleLoc3D[face][1][1].y = -subdiamondLoc3D[face][1].y;
			triangleLoc3D[face][1][2].y = subdiamondLoc3D[face][1].y -
				subdiamondLoc3D[face][2].y;
			triangleLoc3D[face][2][1].x = -subdiamondLoc3D[face][2].x;
			triangleLoc3D[face][2][2].x = subdiamondLoc3D[face][1].x +
				subdiamondLoc3D[face][2].x;
			triangleLoc3D[face][2][1].y = -subdiamondLoc3D[face][2].y;
			triangleLoc3D[face][2][2].y = subdiamondLoc3D[face][1].y +
				subdiamondLoc3D[face][2].y;
			triangleLoc3D[face][3][1].x = subdiamondLoc3D[face][1].x;
			triangleLoc3D[face][3][2].x = subdiamondLoc3D[face][2].x -
				subdiamondLoc3D[face][1].x;
			triangleLoc3D[face][3][1].y = subdiamondLoc3D[face][1].y;
			triangleLoc3D[face][3][2].y = subdiamondLoc3D[face][2].y -
				subdiamondLoc3D[face][1].y;
			for (orient = 0; orient < Dino.MAX_ORIENT; orient++) {
				letter3DList[face][orient] = new Point(
					(2 * triangleLoc3D[face][orient][1].x +
					 triangleLoc3D[face][orient][2].x) / 3,
					(2 * triangleLoc3D[face][orient][1].y +
					 triangleLoc3D[face][orient][2].y) / 3);
				triangleLoc3D[face][orient][3].x =
					-triangleLoc3D[face][orient][1].x -
					triangleLoc3D[face][orient][2].x;
				triangleLoc3D[face][orient][3].y =
					-triangleLoc3D[face][orient][1].y -
					triangleLoc3D[face][orient][2].y;
			}
			for (orient = 1; orient < Dino.MAX_ORIENT; orient++) {
				side = corner;
				corner = mapFrom3D(face, (orient + 1) % Dino.MAX_ORIENT);
				orientTriangle[face][corner][0] = new Point(
					orientTriangle[face][side][0].x +
					orientDiamondLoc3D[0][face][orient].x,
					orientTriangle[face][side][0].y +
					orientDiamondLoc3D[0][face][orient].y);
				orientTriangle[face][corner][1] = new Point(
					orientTriangle[face][side][1].x +
					orientDiamondLoc3D[1][face][orient].x,
					orientTriangle[face][side][1].y +
					orientDiamondLoc3D[1][face][orient].y);
			}
		}
		cubeSize.x = faceSize.x - 2 * delta;
		cubeSize.y = faceSize.y - 2 * delta;
		cubeDiagonal = faceDiagonal - 2 * delta;
		//cubeDiag = faceDiagonal + 2 * delta;
	}

	void narrowSelection() {
		//sets dino.currentFace, dino.currentPosition, dino.currentDirection
		switch (dino.currentDirection) {
		case Dino.TR:
		case Dino.BR:
		case Dino.BL:
		case Dino.TL:
			if (vertical) {
				if (dino.currentFace == 1)
					dino.currentDirection = dino.currentDirection + 2 * Dino.MAX_ORIENT;
				else if (dino.currentFace == 3)
					dino.currentDirection = (dino.currentDirection + 3) % Dino.MAX_ORIENT + 2 * Dino.MAX_ORIENT;
			} else {
				switch (dino.currentFace) {
				case 0:
					dino.currentDirection = (dino.currentDirection + 3) % Dino.MAX_ORIENT + 2 * Dino.MAX_ORIENT;
					break;
				case 4:
					dino.currentDirection = (dino.currentDirection + 2) % Dino.MAX_ORIENT + 2 * Dino.MAX_ORIENT;
					break;
				case 5:
					dino.currentDirection = (dino.currentDirection + 2) % Dino.MAX_ORIENT;
					break;
				default:
					break;
				}
			}
			break;
		case Dino.CCW:
		case Dino.CW:
			/*switch (dino.currentFace) {
			case 3:
			case 4:
				dino.currentDirection = (dino.currentDirection - Dino.CW) % Dino.MAX_ORIENT;
				break;
			case 0:
				dino.currentDirection = (dino.currentDirection - Dino.CW + 1) % Dino.MAX_ORIENT;
				break;
			case 2:
				dino.currentDirection = (dino.currentDirection - Dino.CW + 2) % Dino.MAX_ORIENT;
				break;
			case 1:
			case 5:
				dino.currentDirection = (dino.currentDirection - Dino.CW + 3) % Dino.MAX_ORIENT;
				break;
			default:
				break;
			}*/
			break;
		case Dino.TOP:
		case Dino.RIGHT:
		case Dino.BOTTOM:
		case Dino.LEFT:
			if (vertical) {
				if (dino.currentFace == 1)
					dino.currentDirection = (dino.currentDirection + 3) % Dino.MAX_ORIENT;
				else if (dino.currentFace == 3)
					dino.currentDirection = (dino.currentDirection + 2) % Dino.MAX_ORIENT;
			} else {
				switch (dino.currentFace) {
				case 0:
					dino.currentDirection = dino.currentDirection % Dino.MAX_ORIENT;
					break;
				case 4:
					dino.currentDirection = (dino.currentDirection + 1) % Dino.MAX_ORIENT;
					break;
				case 5:
					dino.currentDirection = (dino.currentDirection + 2) % Dino.MAX_ORIENT +
						2 * Dino.MAX_ORIENT;
					break;
				default:
					break;
				}
			}
			break;
		default:
			dino.currentFace = -1;
		}
/*		if (dino.currentPosition != Dino.MAX_ORIENT) {
			if (dino.currentDirection == Dino.CW)
				dino.currentDirection =
				  (dino.currentPosition + 1) % Dino.MAX_ORIENT;
			else if (dino.currentDirection == Dino.CCW)
				dino.currentDirection =
				  (dino.currentPosition + 3) % Dino.MAX_ORIENT;
			else if (dino.currentDirection < Dino.MAX_ORIENT &&
			  ((dino.currentDirection + dino.currentPosition) % 2 == 0))
				dino.currentFace = -1;
				return;
		} */
	}

	int positionToPiece(int x, int y) { // sets dino.currentPosition
		int u, v, found, face;
		int tempx = x, tempy = y;
		boolean front, tl, ur, ul;
		Point coord;

		if (vertical) {
			tempx -= viewMiddle.x;
			front = (tempy < viewSize.y + puzzleOffset.y);
			if (!front)
				tempy -= viewSize.y;
			tl = (tempy < viewMiddle.y);
			tempy -= viewMiddle.y;
			u = -faceSize.y * tempx + faceDiagonal * tempy;
			v = faceSize.y * tempx + faceDiagonal * tempy;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 0 : 1);
				else
					face = ((ul) ? 1 : 2);
			} else {
				if (tl)
					face = ((ul) ? 4 : 3);
				else
					face = ((ur) ? 3 : 5);
			}
		} else {
			tempy -= viewMiddle.y;
			front = (tempx < viewSize.x + puzzleOffset.x);
			if (!front)
				tempx -= (viewSize.x);
			tl = (tempx < viewMiddle.x);
			tempx -= viewMiddle.x;
			u = -faceSize.x * tempy + faceDiagonal * tempx;
			v = faceSize.x * tempy + faceDiagonal * tempx;
			ur = (u < 0);
			ul = (v < 0);
			if (front) {
				if (tl)
					face = ((ur) ? 1 : 0);
				else
					face = ((ul) ? 0 : 2);
			} else {
				if (tl)
					face = ((ul) ? 3 : 4);
				else
					face = ((ur) ? 4 : 5);
			}
		}
		found = 0;
		coord = diamondOffset3D(face);
		for (int corner = 0; corner < Dino.MAX_ORIENT; corner++) {
			int x1, y1, x2, y2, ok = 0;

			x1 = coord.x + triangleLoc3D[face][corner][1].x;
			y1 = coord.y + triangleLoc3D[face][corner][1].y;
			x2 = x1 + triangleLoc3D[face][corner][2].x;
			y2 = y1 + triangleLoc3D[face][corner][2].y;
			if ((x1 - coord.x) * (y2 - coord.y) <=
					(y1 - coord.y) * (x2 - coord.x)) {
				if ((x1 - coord.x) * (y - coord.y) <=
						(y1 - coord.y) * (x - coord.x))
					ok++;
			} else if ((x1 - coord.x) * (y - coord.y) >=
					(y1 - coord.y) * (x - coord.x))
				ok++;
			if ((x2 - coord.x) * (y1 - coord.y) <=
					(y2 - coord.y) * (x1 - coord.x)) {
				if ((x2 - coord.x) * (y - coord.y) <=
						(y2 - coord.y) * (x - coord.x))
					ok++;
			} else if ((x2 - coord.x) * (y - coord.y) >=
					(y2 - coord.y) * (x - coord.x))
				ok++;
			if (ok == 2) {
				dino.currentPosition = mapFrom3D(face, corner);
				found++;
			}
		}
		if (found != 1)
			return -1;
		return face;
	}

	void resizePuzzle() {
		coreSize = new Point(dino.getSize().width,
			dino.getSize().height);
		delta = 4;
		vertical = (coreSize.y >= coreSize.x);
		if (vertical) {
			cubeSize = new Point(coreSize.x,
				coreSize.y / MAX_VIEWS);
			if (cubeSize.x >= divide(cubeSize.y)) {
				cubeSize.y = java.lang.Math.max((cubeSize.y -
					3 * delta) / 2 - delta - 2, 0);
				cubeSize.x = divide(cubeSize.y);
			} else {
				cubeSize.x = java.lang.Math.max((cubeSize.x -
					2 * delta - 7) / 2 - delta, 0);
				cubeSize.y = multiply(cubeSize.x);
			}
			cubeDiagonal = cubeSize.x / 2;
			faceSize = new Point(
				java.lang.Math.max(cubeSize.x + 2 * delta + 1, 1),
				java.lang.Math.max(cubeSize.y + 2 * delta + 1, 1));
			faceDiagonal = faceSize.x / 2;
			viewSize = new Point(2 * faceSize.x + 3,
				2 * faceSize.y + 3);
			puzzleSize = new Point(viewSize.x + 1,
				MAX_VIEWS * viewSize.y + 1);
		} else {
			cubeSize = new Point(coreSize.x / MAX_VIEWS,
				coreSize.y);
			if (cubeSize.y >= divide(cubeSize.x)) {
				cubeSize.x = java.lang.Math.max((cubeSize.x -
					3 * delta) / 2 - delta - 2, 0);
				cubeSize.y = divide(cubeSize.x);
			} else {
				cubeSize.y = java.lang.Math.max((cubeSize.y -
					2 * delta - 7) / 2 - delta, 0);
				cubeSize.x = multiply(cubeSize.y);
			}
			cubeDiagonal = cubeSize.y / 2;
			faceSize = new Point(
				java.lang.Math.max(cubeSize.x + 2 * delta + 1, 1),
				java.lang.Math.max(cubeSize.y + 2 * delta + 1, 1));
			faceDiagonal = faceSize.y / 2;
			viewSize = new Point(2 * faceSize.x + 3,
				2 * faceSize.y + 3);
			puzzleSize = new Point(MAX_VIEWS * viewSize.x + 1,
				viewSize.y + 1);
		}
		puzzleOffset = new Point((coreSize.x - puzzleSize.x) / 2,
			(coreSize.y - puzzleSize.y) / 2);
		resizePieces();
	}

	public void initializePuzzle() {
		resizePuzzle();
	}

	void exposePuzzle() {
		drawFrame(false);
		drawFrame(true);
		dino.drawAllPieces();
	}

	void selectPuzzle(boolean control, int x, int y) {
		if ((dino.currentFace = positionToPiece(x, y)) != -1) { //sets currentPosition
			if (control || dino.practiceChoice || !dino.checkSolved()) {
				selectPaint = true;
				dino.repaint();
				return;
			}
			dino.currentFace = -1;
		}
		dino.currentFace = -1;
	}

	void releasePuzzle(boolean shift, boolean control, boolean alt,
			int x, int y) {
		if (dino.currentFace == -1)
			return;
		dino.nextShift = shift;
		dino.nextControl = control;
		dino.nextAlt = alt;
		int tempPosition = dino.currentPosition;
		dino.nextFace = positionToPiece(x, y); // sets currentPosition
		dino.nextPosition = dino.currentPosition;
		dino.currentPosition = tempPosition;
		releasePaint = true;
		dino.repaint();
	}

	@Override
	public void paint(Graphics g) {
		if (firstPaint) {
			dino.initializePuzzle();
			initializePuzzle();
			firstPaint = false;
			resizePaint = true;
		}
		if (!framePaint && !allPiecesPaint &&
				!selectPaint && !releasePaint &&
				!scanPaint && !stackPaint &&
				randomPaints == 0) {
			resizePaint = true;
		}
		if (resizePaint) {
			resizePuzzle();
			exposePuzzle();
			resizePaint = false;
		}
		if (framePaint) {
			drawFrame(true);
			framePaint = false;
		}
		if (allPiecesPaint) {
			drawAllPieces();
			allPiecesPaint = false;
		}
		if (selectPaint) {
			if (dino.currentFace != -1 && dino.currentPosition != -1)
				drawTriangle(dino.currentFace,
					dino.currentPosition, 1);
			selectPaint = false;
		}
		if (releasePaint) {
			if (dino.currentFace != -1) {
				drawTriangle(dino.currentFace,
					dino.currentPosition, 0);
				dino.selectPieces();
				dino.currentFace = -1;
			}
			releasePaint = false;
		}
		if (scanPaint || stackPaint) {
			if (dino.currentControl) {
				dino.moveControlCb(dino.currentFace,
					dino.currentPosition,
					dino.currentDirection,
					dino.currentStyle);
			} else {
				dino.movePieces(dino.currentFace,
					dino.currentPosition,
					dino.currentDirection,
					dino.currentStyle);
				if (stackPaint && dino.checkSolved()) {
					((DinoFrame) frame).callback(
						DinoInterface.ACTION_SOLVED,
						3);
				}
			}
			dino.currentDirection = dino.currentFace = -1;
			scanPaint = false;
			stackPaint = false;
			try {
				Thread.sleep(dino.delay);
			} catch (InterruptedException e) {
				// ignore
			}
		}
		if (randomPaints > 0) {
			dino.randomizingPieces();
		}
	}

	@Override
	public void update(Graphics g) {
		paint(g); /* no erase */
	}

	public void keyPressed(KeyEvent event) {
		int modifiers = event.getModifiersEx();

		switch (event.getKeyCode()) {
		case KeyEvent.VK_F11:
		case KeyEvent.VK_DIVIDE:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.CCW,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_HOME:
		case KeyEvent.VK_NUMPAD7:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.TL,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_UP:
		case KeyEvent.VK_KP_UP:
		case KeyEvent.VK_NUMPAD8:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.TOP,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_UP:
		case KeyEvent.VK_NUMPAD9:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.TR,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_LEFT:
		case KeyEvent.VK_KP_LEFT:
		case KeyEvent.VK_NUMPAD4:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.LEFT,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_F12:
		case KeyEvent.VK_CLEAR:
		case KeyEvent.VK_NUMPAD5:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.CW,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_RIGHT:
		case KeyEvent.VK_KP_RIGHT:
		case KeyEvent.VK_NUMPAD6:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.RIGHT,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_END:
		case KeyEvent.VK_NUMPAD1:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.BL,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_DOWN:
		case KeyEvent.VK_KP_DOWN:
		case KeyEvent.VK_NUMPAD2:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.BOTTOM,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		case KeyEvent.VK_PAGE_DOWN:
		case KeyEvent.VK_NUMPAD3:
			dino.movePuzzleInput(dino.lastX, dino.lastY,
				Dino.BR,
				(modifiers & InputEvent.SHIFT_DOWN_MASK) != 0,
				(modifiers & InputEvent.CTRL_DOWN_MASK) != 0,
				(modifiers & InputEvent.ALT_DOWN_MASK) != 0);
			break;
		default:
		}
	}
}
