package com.sillycycle.bagleyd.dino;

/*
 * @(#)DinoFrame.java
 *
 * Copyright 1994 - 2023  David A. Bagley, bagleyd AT verizon.net
 *
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * This program is distributed in the hope that it will be "useful",
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Image;
import java.awt.Label;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.sillycycle.bagleyd.util.ArgumentParser;
import com.sillycycle.bagleyd.util.Colour;
import com.sillycycle.bagleyd.util.ColumnLayout;
import com.sillycycle.bagleyd.util.ComponentUtil;
import com.sillycycle.bagleyd.util.Icon;
import com.sillycycle.bagleyd.util.IODialog;
import com.sillycycle.bagleyd.util.OKDialog;
import com.sillycycle.bagleyd.util.Spinner;

/**
 * The <code>DinoFrame</code> class is to be embedded in a Web page
 * or run as an application.
 *
 * @author	David A. Bagley
 * @author	bagleyd AT verizon.net
 * @author	https://www.sillycycle.com/puzzles.html
 */

public class DinoFrame extends JFrame implements ActionListener {
	private static final long serialVersionUID = 42L;
	public static final String NL = System.getProperty("line.separator");
	static final int MIN_SPEED = 1;
	static final int MAX_SPEED = 50;
	static final int NEVER = -1;
	static final String SCOREFILE = "dino.scores";
	static final String[] COMMAND_STRINGS = new String[] {
		"Get Buffer", "Write Buffer", "Show Buffer",
		"Undo", "Redo", "Clear",
		"Practice", "Randomize", "Auto-solve",
		"Orientize",
		"Speed >", "Slow <", "Sound @",
		"Description ?", "Features !", "References ^", "About",
		"Get File", "Write File", "Exit"};
	static final char[] COMMAND_CHARS = new char[] {
		'g', 'w', '*',
		'u', 'r', 'c', 'p', 'z', 's',
		'o', '>', '<', '@',
		'?', '!', '^', 'a', 'G', 'W', 'x'};

	static ArgumentParser argumentParser = null;
	Icon mediumIcon = null;
	Icon largeIcon = null;
	int widthValue = 320;
	int heightValue = 175;
	static int xValue = 0;
	static int yValue = 0;
	String monoString = "RBYGWO";
	Color[] faceColor = new Color[] {
		Color.red, Color.blue, Color.yellow, Color.green,
		Color.white, Color.orange
	};
	Color borderColor = Color.black;
	Color fgColor = Color.cyan;
	Color bgColor = Colour.steelBlue;
	boolean sound = false;
	boolean controlValue = true;
	int modeValue = DinoInterface.DEFAULT_MODE;
	boolean orientValue = DinoInterface.DEFAULT_ORIENT;
	boolean practiceValue = DinoInterface.DEFAULT_PRACTICE;
	int delayValue = 10;
	int speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
	boolean debug = false;
	JPanel topPanel = new JPanel();
	JComboBox<Object> modeComboBox = new JComboBox<>();
	JCheckBox orientCheckBox = new JCheckBox("Oriented");
	JCheckBox practiceCheckBox = new JCheckBox("Practice");
	DinoStack undo = new DinoStack();
	DinoStack redo = new DinoStack();
	Dino puzzle2D, puzzle3D;
	int movesDisplay = 0;
	String recordDisplay = "NEVER";
	String messageDisplay = "Play Dino! (use mouse or keypad)";
	JTextField movesValueLabel = new JTextField(Integer.toString(
		movesDisplay));
	JTextField recordValueLabel = new JTextField(recordDisplay);
	Label messageLabel = new Label(messageDisplay, Label.CENTER);
	// JLabel constantly resizes
	OKDialog aDialog = null;
	StringBuilder stringBuilderSave;
	IODialog ioDialog = null;
	int[][] puzzleRecord = null;
	JMenuBar menuBar;
	JMenuItem detachMenuItem;
	JPanel mainPanel = new JPanel();
	JPanel secondaryPanel = new JPanel();
	JFrame frame;
	Spinner speedChanger = null;
	//Slider speedChanger = null;

	public String writePuzzleString() {
		return this.stringBuilderSave.toString();
	}

	public void getPuzzleString(String s) {
		puzzle2D.stringSave = s;
		puzzle3D.stringSave = s;
	}

	public void getPuzzle() {
		puzzle2D.getPuzzle(false);
		puzzle3D.getPuzzle(true);
	}

	public void getPuzzleFile() {
		puzzle2D.getPuzzleFile(false);
		puzzle3D.getPuzzleFile(true);
	}

	public void writePuzzle() {
		puzzle3D.writePuzzle();
	}

	public void undoPuzzle() {
		puzzle3D.undoPuzzle();
	}

	public void redoPuzzle() {
		puzzle3D.redoPuzzle();
	}

	public void clearPuzzle() {
		puzzle3D.clearPuzzle();
	}

	public void practicePuzzle() {
		puzzle3D.practicePuzzle();
	}

	public void randomizePuzzle() {
		puzzle3D.randomizePuzzle();
	}

	public void solvePuzzle() {
		puzzle3D.solvePuzzle();
	}

	public void orientizePuzzle() {
		puzzle3D.orientizePuzzle();
	}

	public void processKey(char key) {
		puzzle3D.processKey(key);
	}

	public boolean getToggleSound() {
		return sound;
	}

	public void toggleSound() {
		sound = !sound;
	}

	public String getOpt(String string) {
		return argumentParser.getOption(string);
	}

	public int getOpt(String param, int value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Integer.parseInt(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public boolean getOpt(String param, boolean value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			if (string.equalsIgnoreCase("true") ||
					string.equalsIgnoreCase("yes") ||
					string.equalsIgnoreCase("on"))
				return true;
			else if (string.equalsIgnoreCase("false") ||
					string.equalsIgnoreCase("no") ||
					string.equalsIgnoreCase("off"))
				return false;
			return (Integer.parseInt(string) != 0);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public Color getOpt(String param, Color value) {
		String string = getOpt(param);

		if (string == null)
			return value;
		try {
			return Color.decode(string);
		} catch (NumberFormatException e) {
			return value; // ignore garbage
		}
	}

	public void getOpts() {
		String string;

		widthValue = getOpt("windowWidth", widthValue);
		heightValue = getOpt("windowHeight", heightValue);
		xValue = getOpt("windowX", xValue);
		yValue = getOpt("windowY", yValue);
		for (int i = 0; i < faceColor.length; i++)
			faceColor[i] = getOpt("faceColor" + i, faceColor[i]);
		borderColor = getOpt("borderColor", borderColor);
		fgColor = getOpt("fgColor", fgColor);
		bgColor = getOpt("bgColor", bgColor);
		sound = getOpt("sound", sound);
		controlValue = getOpt("control", controlValue);
		modeValue = getOpt("mode", modeValue);
		string = getOpt("both");
		if (string != null)
			modeValue = DinoInterface.BOTH;
		if (modeValue < DinoInterface.PERIOD2 ||
				modeValue > DinoInterface.BOTH)
			modeValue = DinoInterface.DEFAULT_MODE;
		orientValue = getOpt("orient", orientValue);
		practiceValue = getOpt("practice", practiceValue);
		delayValue = getOpt("delay", delayValue);
	}

	public void showBuffer() {
		if (ioDialog != null) {
			ioDialog.setVisible(true);
		}
	}

	public void showMessage(String msg) {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Dino", msg, "OK",
				mediumIcon, false);
			aDialog.setBounds(this.getSize().width / 2,
				this.getSize().height / 2, 350, 120);
			aDialog.setVisible(true);
		}
	}

	public void descriptionHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Description",
"Dino - the original puzzle has period 3 turning (i.e. the" + NL +
"points turn with 120 degree intervals) is called a" + NL +
"Triangle - 4 Cube - Dinosaur with colored dinosaurs on it," + NL +
"(or Triangle - 4 Cube - 6 Colors with solid colors, (a 4" + NL +
"color and a 2 color are also available with point oriented" + NL +
"coloring)) manufactured by The International Puzzle &" + NL +
"Games (Great Animation International Corporation)." + NL +
"The period 2 turning (i.e. the edges turn with 180 degree" + NL +
"intervals) was suggested by Derek Bosch <bosch@sgi.com>." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void featuresHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"Features",
"Make sure you randomize the puzzle first before trying to move the" + NL +
"pieces for the first time." + NL +
" " + NL +
"Press \"mouse-left\" button to move a piece.  Release \"mouse-left\"" + NL +
"button on another piece on the same face.  (Clicking on the" + NL +
"diamonds are ignored).  The pieces will then turn towards where the" + NL +
"mouse button was released." + NL +
" " + NL +
"Click \"mouse-right\" button, or press \"Z\" or \"z\" keys, to" + NL +
"randomize the puzzle (this must be done first to set a new record)." + NL +
"One must double click on \"mouse-right\" if the puzzle is currently" + NL +
"being worked on." + NL +
" " + NL +
"Press \"g\" key to get (read) a saved puzzle to buffer." + NL +
" " + NL +
"Press \"w\" key to save (write) a puzzle to buffer." + NL +
" " + NL +
"Press \"G\" key to get (read) a saved puzzle to file." + NL +
" " + NL +
"Press \"W\" key to save (write) a puzzle to file." + NL +
" " + NL +
"Press \"U\" or \"u\" keys to undo a move." + NL +
" " + NL +
"Press \"R\" or \"r\" keys to redo a move." + NL +
" " + NL +
"Press \"C\" or \"c\" keys to clear the puzzle." + NL +
" " + NL +
"Press \"P\" or \"p\" keys, to toggle the practice mode (in practice" + NL +
"mode the record should say \"practice\").  This is good for" + NL +
"learning moves and experimenting." + NL +
" " + NL +
"Press \"S\" or \"s\" keys to start auto-solver." + NL +
" " + NL +
"Press \"O\" or \"o\" keys to toggle the orient mode.  (Default is off)." + NL +
"One has to orient the faces in orient mode, besides getting all" + NL +
"the faces to be the same color.  To do this one has to get the" + NL +
"lines to be oriented in the same direction.  This does add complexity," + NL +
"(ever so slightly, there are only 2 possibilities), so there are 2" + NL +
"sets of records." + NL +
" " + NL +
"Press \"2\", \"3\", \"B\", or \"b\" keys (not the keypad 2, 3)" + NL +
"to change modes to Period 2, Period 3, or Both." + NL +
" " + NL +
"Press \">\" or \".\" keys to speed up the movement of the pieces." + NL +
" " + NL +
"Press \"<\" or \",\" keys to slow up the movement of the pieces." + NL +
" " + NL +
"Press \"@\" key to toggle the sound." + NL +
" " + NL +
"Press \"Esc\" key to hide program." + NL +
" " + NL +
"Press \"Q\", \"q\", or \"CTRL-C\" keys to kill program if an application." + NL +
" " + NL +
"Use the key pad or arrow keys to move without the mouse." + NL +
"Key pad (better with Num Lock on) is defined for Dino2D as:" + NL +
"   /    Counterclockwise" + NL +
" " + NL +
"7 8 9  Upper Left, Up, Upper Right" + NL +
"   ^" + NL +
"4<5>6  Left, Clockwise, Right" + NL +
"   v" + NL +
"1 2 3  Lower Left, Down, Lower Right" + NL +
" " + NL +
"Note: Top, Left, Right, and Bottom only work when the control key is" + NL +
"pressed and there is no analog for Dino3D." + NL +
" " + NL +
"Key pad for Dino3D, use must use your intuition (is this a cop" + NL +
"out or what?).  The key pad is defined differently depending on" + NL +
"which side of the cube your mouse is pointing at.  One thing that" + NL +
"stays the same is \"5\" is Clockwise and \"/\" is Counterclockwise." + NL +
" " + NL +
"Use the alt key and the left mouse button, keypad, or arrow" + NL +
"keys to move the center of the cube.  The 2 opposite corners" + NL +
"do not move." + NL +
"Use the shift keys to access \"Period 3\" turns from \"Both\" mode," + NL +
"otherwise it assumes \"Period 2\" turning.  Faces and points turn in" + NL +
"\"Period 3\" and edges (2 points) turn in \"Period 2\"." + NL +
"Use the control key and the left mouse button, keypad, or arrow" + NL +
"keys to move the whole puzzle.  This is not recorded as a turn." + NL,
				"OK", largeIcon, true);
			aDialog.setBackground(Color.white);
			aDialog.setVisible(true);
		}
	}

	public void referencesHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"References",
"International Puzzles & Games Catalog.",
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public void aboutHelp() {
		if (aDialog == null || !OKDialog.getActive()) {
			aDialog = new OKDialog(ComponentUtil.findJFrame(this),
				"About",
"Java DinoFrame Version 8.7.3" + NL +
"Send bugs (reports or fixes) to the author:" +
"David Bagley <bagleyd AT verizon.net>" + NL +
"The latest version is at: " +
"https://www.sillycycle.com/puzzles.html" + NL,
				"OK", mediumIcon, false);
			aDialog.setVisible(true);
		}
	}

	public Icon getIcon(String string, int sx, int sy) {
		URL url = DinoFrame.class.getResource("/" + string);

		if (url == null)
			return null;
		return new Icon(Toolkit.getDefaultToolkit().
			getImage(url), sx, sy);
	}

	// TODO Make more reliable, also should cache
	public void playClip(String string) {
		URL url = DinoFrame.class.getResource("/" + string);

		if (url == null)
			return;
		try {
			AudioInputStream audioInputStream =
				AudioSystem.getAudioInputStream(url);
			DataLine.Info info = new DataLine.Info(Clip.class,
				audioInputStream.getFormat());
			Clip clip = (Clip)AudioSystem.getLine(info);
			clip.open(audioInputStream);
			clip.start();
		} catch (IOException | LineUnavailableException | UnsupportedAudioFileException e) {
			e.printStackTrace();
		}
	}

	public void playMoveAudio() {
		playClip(DinoInterface.SOUNDS[DinoInterface.Sounds.MOVE.ordinal()]);
	}

	@Override
	public void setSize(int width, int height) {
		if (puzzle2D == null || puzzle3D == null)
			return;
		puzzle2D.setBounds(0, 0, width, height / 2);
		puzzle3D.setBounds(0, 0, width, height - height / 2);
		validate();
		/*repaint();*/ // causes flashing
	}

	@Override
	public void paint(Graphics g) {
		setSize(secondaryPanel.getSize().width,
			secondaryPanel.getSize().height - topPanel.getSize().height);
		super.paint(g);
	}

	public String readUserScore() {
		StringBuilder scoreString = new StringBuilder();

		for (int i = 0; i < puzzleRecord.length; i++) {
			for (int orient = 0; orient < puzzleRecord.length; orient++) {
				scoreString.append(puzzleRecord[i][orient]);
				if (orient < puzzleRecord[i].length - 1)
					scoreString.append(" ");
			}
			if (i < puzzleRecord.length - 1)
				scoreString.append(NL);
		}
		return scoreString.toString();
	}

	public void writeScore() {
		try (PrintWriter pr = new PrintWriter(new FileWriter(
				SCOREFILE, false))) {
			pr.println(readUserScore());
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	public void initRecords(String cookieValue, boolean store) {
		String[] st = null;
		int field = 0;

		if (cookieValue != null)
			st = cookieValue.split("[+/\\s]+");
		if (debug)
			System.out.println("initRecords: " + cookieValue);
		puzzleRecord = new int[DinoInterface.MAX_MODES][2];
		for (int i = 0; i < puzzleRecord.length; i++) {
		    for (int orient = 0; orient < puzzleRecord[i].length; orient++)
			if (st == null || st.length <= field) {
				//build it from the array
				puzzleRecord[i][orient] =
					DinoScores.PUZZLE_SCORES[i][orient];
			} else {
				puzzleRecord[i][orient] =
					Integer.parseInt(st[field++]);
			}
		}
		if (store && st == null)
			writeScore(); // if no cookie, then make one...
	}

	public void readScoreRecords() {
		try (BufferedReader br = new BufferedReader(
				new FileReader(SCOREFILE))) {
			StringBuilder scoreData = new StringBuilder();
			String lineData;

			while ((lineData = br.readLine()) != null) {
				scoreData.append(lineData);
				scoreData.append(NL);
			}
			initRecords(scoreData.toString(), true);
		} catch (Exception e) {
			//e.printStackTrace();
			showMessage("Cannot open " + SCOREFILE);
			initRecords(null, true);
		}
	}

	public String readTextArea() {
		return ioDialog.getText();
	}

	public void writeTextArea(String s) {
		ioDialog.setText(s);
	}

	public void shuffleDown() {
		ComponentUtil.findJFrame(this).toBack();
	}

	public void quit() {
		System.exit(0);
	}

	void printRecord(int mode,
			boolean orient, boolean practice) {
		int i = mode - DinoInterface.PERIOD2;
		int j = (orient) ? 1 : 0;

		if (puzzleRecord == null)
			return;
		if (practice) {
			recordValueLabel.setText("PRACTICE");
		} else if (puzzleRecord[i][j] <= NEVER) {
			recordValueLabel.setText("NEVER");
		} else {
			recordValueLabel.setText(Integer.toString(
				puzzleRecord[i][j]));
		}
	}

	boolean handleSolved(int counter, boolean orient,
			boolean practice) {
		int i = modeValue - DinoInterface.PERIOD2;
		int j = (orient) ? 1 : 0;
		int notJ = (orient) ? 0 : 1;

		if (puzzleRecord != null &&
				(counter < puzzleRecord[i][j] ||
				puzzleRecord[i][j] <= NEVER)) {
			readScoreRecords(); /* Maybe its been updated by another */
			puzzleRecord[i][j] = counter;
			if (orient && (counter < puzzleRecord[i][notJ] ||
					puzzleRecord[i][notJ] <= NEVER)) {
				puzzleRecord[i][notJ] = counter;
			}
			writeScore();
			printRecord(modeValue, orient, practice);
			return true;
		}
		return false;
	}

	public void initScore() {
		if (puzzleRecord == null) {
			readScoreRecords();
			printRecord(modeValue, orientValue, practiceValue);
		}
	}

	class SpeedChangeListener implements ChangeListener {
		@Override
		public void stateChanged(ChangeEvent event) {
			int oldValue = speedValue;

			speedValue = speedChanger.getValue();
			if (oldValue != speedValue) {
				puzzle2D.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
				puzzle3D.setDelay(MAX_SPEED + MIN_SPEED - 1 -
					speedValue);
			}
		}
	}

	class ComboBoxActionListener implements ActionListener {
		@Override
		public void actionPerformed(ActionEvent event) {
			int modeIndex = modeComboBox.getSelectedIndex()
				+ DinoInterface.PERIOD2;

			if (modeIndex != -1 && modeValue != modeIndex) {
				modeValue = modeIndex;
				puzzle2D.setPuzzlePeriod(modeValue);
				puzzle3D.setPuzzlePeriod(modeValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(modeValue,
					orientValue, practiceValue);
			}
		}
	}

	class OrientItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = orientValue;

			orientValue = orientCheckBox.isSelected();
			if (oldValue != orientValue) {
				puzzle2D.setPuzzleOrient(orientValue);
				puzzle3D.setPuzzleOrient(orientValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				printRecord(modeValue,
					orientValue, practiceValue);
			}
		}
	}

	class PracticeItemListener implements ItemListener {
		@Override
		public void itemStateChanged(ItemEvent event) {
			boolean oldValue = practiceValue;

			practiceValue = practiceCheckBox.isSelected();
			if (oldValue != practiceValue) {
				puzzle2D.setPuzzlePractice(practiceValue);
				puzzle3D.setPuzzlePractice(practiceValue);
				movesDisplay = 0;
				movesValueLabel.setText(Integer.toString(
					movesDisplay));
				if (!practiceValue)
					messageLabel.setText(
						"Randomize to start");
				printRecord(modeValue,
					orientValue, practiceValue);
			}
		}
	}

	@Override
	public void actionPerformed(ActionEvent event) {
		String command = event.getActionCommand();

		for (int opt = 0; opt < COMMAND_STRINGS.length; opt++) {
			if (command.equals(COMMAND_STRINGS[opt])) {
				processKey(COMMAND_CHARS[opt]);
				return;
			}
		}
	}

	public void callback(int reason, int dim,
			int face, int position, int direction, int style) {
		messageDisplay = "                                                                               ";
		switch(reason) {
		case DinoInterface.ACTION_MOVED:
			movesDisplay++;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction,
					style);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction,
					style);
			}
			break;
		case DinoInterface.ACTION_CONTROL:
			if (dim == 2) {
				puzzle3D.movePieces(face, position, direction,
					style);
			} else {
				puzzle2D.movePieces(face, position, direction,
					style);
			}
			break;
		case DinoInterface.ACTION_UNDO:
			movesDisplay--;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction,
					style);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction,
					style);
			}
			break;
		case DinoInterface.ACTION_REDO:
			movesDisplay++;
			if (dim == 2) {
				puzzle3D.setStart(true);
				puzzle3D.movePieces(face, position, direction,
					style);
			} else {
				puzzle2D.setStart(true);
				puzzle2D.movePieces(face, position, direction,
					style);
			}
			break;
		case DinoInterface.ACTION_SPEED:
			{
				int oldDelay, delay;

				if (dim == 2) {
					delay = puzzle2D.getDelay();
				} else {
					delay = puzzle3D.getDelay();
				}
				oldDelay = delay;
				if (delay > MAX_SPEED - MIN_SPEED)
					delay = MAX_SPEED - MIN_SPEED;
				if (delay < 0)
					delay = 0;
				if (delay != oldDelay) {
					if (dim == 2) {
						puzzle2D.setDelay(delay);
					} else {
						puzzle3D.setDelay(delay);
					}
				}
				if (speedChanger != null)
					speedChanger.setValue(MAX_SPEED +
						MIN_SPEED - 1 - delay);
			}
			return;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public void callback(int reason, int dim) {
		messageDisplay = "                                                                                ";
		switch(reason) {
		case DinoInterface.ACTION_CLEAR:
			if (dim == 2) {
				puzzle3D.resetPieces();
				puzzle3D.drawAllPieces();
			} else {
				puzzle2D.resetPieces();
				puzzle2D.drawAllPieces();
			}
			movesDisplay = 0;
			break;
		case DinoInterface.ACTION_RESTORE:
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			movesDisplay = 0;
			break;
		case DinoInterface.ACTION_RESET:
			movesDisplay = 0;
			break;
		case DinoInterface.ACTION_AMBIGUOUS:
			messageDisplay = "Ambiguous move";
			break;
		case DinoInterface.ACTION_ILLEGAL:
			if (practiceValue ||
					puzzle2D.checkStarted() ||
					puzzle3D.checkStarted())
				messageDisplay = "Illegal move";
			else
				messageDisplay = "Randomize to start";
			break;
		case DinoInterface.ACTION_MOVED:
			movesDisplay++;
			puzzle2D.setStart(true);
			puzzle3D.setStart(true);
			break;
		case DinoInterface.ACTION_SOLVED:
			if (practiceValue)
				movesDisplay = 0;
			else {
				boolean cheat;

				if (dim == 2) {
					cheat = puzzle2D.getCheat();
				} else {
					cheat = puzzle3D.getCheat();
				}
				if (cheat)
					messageDisplay = "No cheating!!"; //username
				else if (handleSolved(movesDisplay,
						orientValue, practiceValue))
					messageDisplay = "Best solution, congratulations!!"; //username
				else
					messageDisplay = "Solved!";
			}
			if (dim == 2) {
				puzzle3D.setStart(false);
			} else {
				puzzle2D.setStart(false);
			}
			//showMessage(messageDisplay);
			break;
		case DinoInterface.ACTION_COMPUTED:
			if (dim == 2) {
				puzzle3D.setStart(false);
			} else {
				puzzle2D.setStart(false);
			}
			break;
		case DinoInterface.ACTION_CHEAT:
			if (dim == 2) {
				puzzle3D.setCheat(puzzle2D.getCheat());
			} else {
				puzzle2D.setCheat(puzzle3D.getCheat());
			}
			break;
		case DinoInterface.ACTION_PRACTICE:
			movesDisplay = 0;
			practiceValue = !practiceValue;
			if (!practiceValue)
				messageDisplay = "Randomize to start";
			printRecord(modeValue, orientValue, practiceValue);
			puzzle2D.setPuzzlePractice(practiceValue);
			puzzle3D.setPuzzlePractice(practiceValue);
			practiceCheckBox.setSelected(practiceValue);
			puzzle2D.setStart(false);
			puzzle3D.setStart(false);
			break;
		case DinoInterface.ACTION_RANDOMIZE:
			movesDisplay = 0;
			practiceValue = false;
			puzzle2D.setPuzzlePractice(practiceValue);
			puzzle3D.setPuzzlePractice(practiceValue);
			puzzle2D.setStart(false);
			puzzle3D.setStart(false);
			break;
		case DinoInterface.ACTION_ORIENTIZE:
			movesDisplay = 0;
			orientValue = !orientValue;
			printRecord(modeValue, orientValue, practiceValue);
			puzzle2D.setPuzzleOrient(orientValue);
			puzzle3D.setPuzzleOrient(orientValue);
			orientCheckBox.setSelected(orientValue);
			break;
		case DinoInterface.ACTION_PERIOD2:
		case DinoInterface.ACTION_PERIOD3:
		case DinoInterface.ACTION_BOTH:
			movesDisplay = 0;
			modeValue = reason -
				DinoInterface.ACTION_PERIOD2 +
				DinoInterface.PERIOD2;
			printRecord(modeValue, orientValue, practiceValue);
			puzzle2D.setPuzzlePeriod(modeValue);
			puzzle3D.setPuzzlePeriod(modeValue);
			modeComboBox.setSelectedItem(
				DinoInterface.modeStrings[modeValue -
				DinoInterface.PERIOD2]);
			break;
		default:
		}
		messageLabel.setText(messageDisplay);
		movesValueLabel.setText(Integer.toString(movesDisplay));
	}

	public DinoFrame(String title) {
		/*JPanel info1Panel = new JPanel();
		JPanel leftPanel = new JPanel();
		JPanel rightPanel = new JPanel();*/
		JPanel startPanel = new JPanel();
		JPanel info2Panel = new JPanel();
		JPanel movesPanel = new JPanel();
		JPanel recordPanel = new JPanel();
		JPanel checkBoxPanel = new JPanel();
		JPanel comboBoxPanel = new JPanel();
		JPanel canvasPanel = new JPanel();
		JPanel canvas2Panel = new JPanel();
		JPanel canvas3Panel = new JPanel();
		int mode;

		try {
			mediumIcon = getIcon(DinoInterface.ICONS[DinoInterface.Icons.X32.ordinal()], 32, 32);
			largeIcon = getIcon(DinoInterface.ICONS[DinoInterface.Icons.X48.ordinal()], 48, 48);
		} catch (Exception e) {
			largeIcon = null;
		}
		ioDialog = new IODialog(ComponentUtil.findJFrame(this),
			DinoInterface.TITLE + ": IO", "Puzzle Position",
			"DONE");
		readScoreRecords();
		getOpts();
		speedValue = MAX_SPEED + MIN_SPEED - 1 - delayValue;
		mainPanel.setLayout(new BorderLayout());
		mainPanel.setBackground(Color.white);
		if (controlValue) {
			menuBar = new JMenuBar();
			JMenu menu;
			JMenuItem menuItem;

			JPopupMenu.setDefaultLightWeightPopupEnabled(false);
			menu = new JMenu("File");
			menuItem = new JMenuItem("Get Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_G);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write Buffer");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_W);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Show Buffer");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Get File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Write File");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menu.addSeparator();
			menuItem = new JMenuItem("Exit");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_X);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Play");
			menuItem = new JMenuItem("Undo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_U);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Redo");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_R);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Clear");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_C);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Practice");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_P);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Randomize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_Z);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Auto-solve");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_S);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Orientize");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_O);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Speed >");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Slow <");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Sound @");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			menu = new JMenu("Help");
			menuItem = new JMenuItem("Description ?");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("Features !");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("References ^");
			menu.add(menuItem);
			menuItem.addActionListener(this);
			menuItem = new JMenuItem("About");
			menu.add(menuItem);
			menuItem.setMnemonic(KeyEvent.VK_A);
			menuItem.addActionListener(this);
			menuBar.add(menu);
			//setJMenuBar(menuBar); // needs to be in a panel
			mainPanel.add(menuBar, BorderLayout.PAGE_START);
			/*leftPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			leftPanel.add(new JLabel("Move"));
			rightPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
			rightPanel.add(new JLabel("Randomize"));
			leftPanel.add(getIcon("mouse-l.png", 16, 16));
			rightPanel.add(getIcon("mouse-r.png", 16, 16));
			info1Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 0));
			info1Panel.add(leftPanel);
			info1Panel.add(rightPanel);*/
			movesPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			movesPanel.add(new JLabel("Moves"));
			movesValueLabel.setColumns(5);
			movesValueLabel.setEditable(false);
			movesPanel.add(movesValueLabel);
			recordPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 0));
			recordPanel.add(new JLabel("Record"));
			recordValueLabel.setColumns(8);
			recordValueLabel.setEditable(false);
			recordPanel.add(recordValueLabel);
			info2Panel.setLayout(new FlowLayout(FlowLayout.LEFT, 15, 5));
			info2Panel.add(movesPanel);
			info2Panel.add(recordPanel);
			comboBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			for (mode = 0; mode < DinoInterface.MAX_MODES; mode++) {
				modeComboBox.addItem(
					DinoInterface.modeStrings[mode]);
			}
			modeComboBox.setSelectedItem(DinoInterface.
				modeStrings[modeValue - DinoInterface.PERIOD2]);
			comboBoxPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			comboBoxPanel.add(modeComboBox);
			comboBoxPanel.add(new JLabel("Turning Mode"));
			checkBoxPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 5, 5));
			checkBoxPanel.add(orientCheckBox);
			orientCheckBox.setSelected(orientValue);
			checkBoxPanel.add(practiceCheckBox);
			practiceCheckBox.setSelected(practiceValue);
			if (practiceValue)
				recordValueLabel.setText("PRACTICE");
			speedChanger = new Spinner("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			/*speedChanger = new Slider("Animation Speed",
				speedValue, MIN_SPEED,
				((speedValue > MAX_SPEED) ? speedValue : MAX_SPEED));
			speedChanger.setWidth(150);*/
		        checkBoxPanel.add(speedChanger);
			topPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.CENTER));
			/*topPanel.add(info1Panel);*/
			topPanel.add(info2Panel);
			topPanel.add(comboBoxPanel);
			topPanel.add(checkBoxPanel);
			topPanel.add(messageLabel);
			startPanel.setLayout(new ColumnLayout(0, 0, 0,
				ColumnLayout.LEFT));
			startPanel.add(topPanel);
			secondaryPanel.setLayout(new BorderLayout());
			secondaryPanel.add(startPanel, BorderLayout.PAGE_START);
		}
		puzzle2D = new Dino(this, 2,
			modeValue, orientValue, practiceValue,
			false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3], faceColor[4], faceColor[5],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		puzzle3D = new Dino(this, 3,
			modeValue, orientValue, practiceValue,
			false, false, monoString,
			faceColor[0], faceColor[1], faceColor[2],
			faceColor[3], faceColor[4], faceColor[5],
			borderColor, fgColor, bgColor,
			MAX_SPEED + MIN_SPEED - 1 - speedValue);
		canvas2Panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvas3Panel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		canvas2Panel.add(puzzle2D);
		canvas3Panel.add(puzzle3D);
		JPanel extraPanel = new JPanel();
		canvasPanel.setLayout(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.anchor = GridBagConstraints.FIRST_LINE_START;
		c.gridx = 0;
		c.gridy = 0;
		canvasPanel.add(canvas2Panel, c);
		c.gridx = 0;
		c.gridy = 1;
		canvasPanel.add(canvas3Panel, c);
		extraPanel.setLayout(new ColumnLayout(0, 0, 0,
			ColumnLayout.LEFT));
		extraPanel.add(canvasPanel);
		secondaryPanel.add(extraPanel, BorderLayout.CENTER);
		mainPanel.add(secondaryPanel, BorderLayout.CENTER);
		getContentPane().add(mainPanel, BorderLayout.CENTER);
		if (controlValue) {
			printRecord(modeValue, orientValue, practiceValue);
			modeComboBox.addActionListener(new ComboBoxActionListener());
			orientCheckBox.addItemListener(new OrientItemListener());
			practiceCheckBox.addItemListener(new PracticeItemListener());
			speedChanger.addChangeListener(new SpeedChangeListener());
		}
		canvas2Panel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle2D.requestFocus();
			}
		});
		canvas3Panel.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseEntered(MouseEvent event) {
				puzzle3D.requestFocus();
			}
		});
		setTitle(title);
		/* puzzle2D.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle2D.setSize(widthValue, heightValue);
		/* puzzle3D.setSize(getSize().width,
			getSize().height - topPanel.getSize().height); */
		puzzle3D.setSize(widthValue, heightValue);
	}

	public static void main(String[] args) {
		String laf = UIManager.getCrossPlatformLookAndFeelClassName();

		try {
			UIManager.setLookAndFeel(laf);
		} catch (ClassNotFoundException | InstantiationException | IllegalAccessException | UnsupportedLookAndFeelException e) {
			e.printStackTrace();
			System.exit(1);
		}
		List<Image> iconList = new ArrayList<>();
		for (String icon : DinoInterface.ICONS) {
			URL url = DinoFrame.class.getResource("/" + icon);
			if (url != null)
				iconList.add(Toolkit.getDefaultToolkit().getImage(url));
		}

		argumentParser = new ArgumentParser(args);
		if (argumentParser.hasOption("?")) {
			System.out.println("usage:" + NL +
"\t[-windowWidth={int}] [-windowHeight={int}]" + NL +
"\t[-windowX={int}] [-windowY={int}]" + NL +
"\t[-faceColor{0|1|2|3|4|5}={int}]" + NL +
"\t[-borderColor={int}] [-fgColor={int}] [-bgColor={int}]" + NL +
"\t[-sound={bool}] [-control={bool}] [-{mode={int} | -both]" + NL +
"\t[-{orient={bool} [-practice={bool}] [-delay={int}]");
			return;
		}
		DinoFrame frame = new DinoFrame(DinoInterface.TITLE);
		frame.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				System.exit(0);
			}
		});
		if (!iconList.isEmpty()) {
			frame.setIconImages(iconList);
		}
		frame.pack();
		frame.puzzle3D.requestFocusInWindow();
		frame.setLocation(DinoFrame.xValue, DinoFrame.yValue);
		frame.setVisible(true);
	}
}
